import Foundation


class RegisterDomainDetailsViewModel {

    typealias Localized = RegisterDomainDetails.Localized
    typealias CodeNameTuple = (code: String, name: String)

    enum Constant {
        static let phoneNumberCountryCodePrefix = "+"
        static let phoneNumberConnectingChar: Character = "."
        static let maxExtraAddressLine = 5
    }

    enum Change: Equatable {
        case rowValidated(context: ValidationRule.Context, indexPath: IndexPath, isValid: Bool, errorMessage: String?)
        case sectionValidated(context: ValidationRule.Context, sectionIndex: Int, isValid: Bool)
        case formValidated(context: ValidationRule.Context, isValid: Bool)

        case multipleChoiceRowValueChanged(indexPath: IndexPath)

        case addNewAddressLineEnabled(indexPath: IndexPath)
        case addNewAddressLineReplaced(indexPath: IndexPath)

        case checkMarkRowsUpdated(sectionIndex: Int)

        case registerSucceeded(domain: String)

        case loading(Bool)

        case remoteValidationFinished

        case prefillSuccess
        case prefillError(message: String)

        case unexpectedError(message: String)
    }

    enum SectionIndex: Int {
        case privacyProtection
        case contactInformation
        case phone
        case address
    }

    var onChange: ((Change) -> Void)?

    var registerDomainDetailsService: RegisterDomainDetailsServiceProxyProtocol = RegisterDomainDetailsServiceProxy()

    let domain: DomainSuggestion
    let site: JetpackSiteRef
    let domainPurchasedCallback: ((String) -> Void)

    private(set) var addressSectionIndexHelper = CellIndex.AddressSectionIndexHelper()
    private(set) var states: [CodeNameTuple]?
    private(set) var countries: [CodeNameTuple]?

    var countryNames: [String] {
        return countries?.map { $0.name } ?? []
    }
    var stateNames: [String] {
        return states?.map { $0.name } ?? []
    }

    private(set) var isLoading: Bool = false {
        didSet {
            onChange?(.loading(isLoading))
        }
    }

    init(site: JetpackSiteRef, domain: DomainSuggestion, domainPurchasedCallback: @escaping ((String) -> Void)) {
        self.site = site
        self.domain = domain
        self.domainPurchasedCallback = domainPurchasedCallback
        manuallyTriggerValidation()
    }

    lazy var sectionChangeHandler: ((Section.Change) -> Void)? = { [weak self] (change) in
        guard let strongSelf = self else { return }

        switch change {
        case let .rowValidation(context, indexPath, isValid, errorMessage):
            strongSelf.onChange?(.rowValidated(context: context,
                                                indexPath: indexPath,
                                                isValid: isValid,
                                                errorMessage: errorMessage))
        case let .sectionValidation(context, sectionIndex, isSectionValid):
            strongSelf.onChange?(.sectionValidated(context: context, sectionIndex: sectionIndex.rawValue, isValid: isSectionValid))
            strongSelf.onChange?(.formValidated(context: context, isValid: strongSelf.isValid(inContext: context)))
        case let .checkMarkRowsUpdated(sectionIndex):
            strongSelf.onChange?(.checkMarkRowsUpdated(sectionIndex: sectionIndex.rawValue))
        case let .multipleChoiceRowValueChanged(indexPath, row):
            if row.editingStyle == .multipleChoice {
                strongSelf.onChange?(.multipleChoiceRowValueChanged(indexPath: indexPath))
            }
        }
    }

    lazy var sections = [
        Section(
            rows: RegisterDomainDetailsViewModel.privacyProtectionRows,
            sectionIndex: .privacyProtection,
            onChange: sectionChangeHandler
        ),
        Section(
            rows: RegisterDomainDetailsViewModel.contactInformationRows,
            sectionIndex: .contactInformation,
            onChange: sectionChangeHandler
        ),
        Section(
            rows: RegisterDomainDetailsViewModel.phoneNumberRows,
            sectionIndex: .phone,
            onChange: sectionChangeHandler
        ),
        Section(
            rows: RegisterDomainDetailsViewModel.addressRows,
            sectionIndex: .address,
            onChange: sectionChangeHandler
        )
    ]

    func enableAddAddressRow() {
        if !addressSectionIndexHelper.isAddNewAddressVisible
            && Constant.maxExtraAddressLine > addressSectionIndexHelper.addNewAddressIndex {
            addressSectionIndexHelper.isAddNewAddressVisible = true
            sections[SectionIndex.address.rawValue].insert(
                .addAddressLine(
                    title: String(
                        format: Localized.Address.addNewAddressLine,
                        "\(addressSectionIndexHelper.addNewAddressIndex + 1)"
                    )
                ),
                at: addressSectionIndexHelper.addNewAddressIndex
            )
            onChange?(
                .addNewAddressLineEnabled(
                    indexPath: IndexPath(
                        row: addressSectionIndexHelper.addNewAddressIndex,
                        section: SectionIndex.address.rawValue
                    )
                ))
        }
    }

    func replaceAddNewAddressLine() {
        if addressSectionIndexHelper.isAddNewAddressVisible {
            addressSectionIndexHelper.addNewAddressField()
            addressSectionIndexHelper.isAddNewAddressVisible = false
            sections[SectionIndex.address.rawValue].remove(at: addressSectionIndexHelper.addNewAddressIndex)
            sections[SectionIndex.address.rawValue].insert(
                RegisterDomainDetailsViewModel.addressLine(
                    row: addressSectionIndexHelper.addNewAddressIndex
                ),
                at: addressSectionIndexHelper.addNewAddressIndex
            )
            onChange?(.addNewAddressLineReplaced(indexPath: IndexPath(
                row: addressSectionIndexHelper.addNewAddressIndex,
                section: SectionIndex.address.rawValue
            )))
        }
    }

    func updateValue<T>(_ value: T?, at indexPath: IndexPath) {
        sections[indexPath.section].updateValue(value, at: indexPath.row)
    }

    func isValid(inContext context: ValidationRule.Context) -> Bool {
        for section in sections {
            if !section.isValid(inContext: context) {
                return false
            }
        }
        return true
    }

    private func manuallyTriggerValidation() {
        sections.forEach {
            $0.triggerValidation()
        }
    }

    func register() {
        let privacyEnabled = privacySectionSelectedItem() == CellIndex.PrivacyProtection.privately

        isLoading = true
        validateRemotely(successCompletion: { [weak self] in
            guard let strongSelf = self else {
                self?.isLoading = false
                return
            }


            WPAnalytics.track(.automatedTransferCustomDomainContactInfoValidated)

            // This is a bit of a callback hell, but our services aren't super mobile friendly.
            // We'll manage.

            // First step is to create a cart.
            strongSelf.registerDomainDetailsService.createShoppingCart(siteID: strongSelf.site.siteID,
                                                                       domainSuggestion: strongSelf.domain,
                                                                       privacyProtectionEnabled: privacyEnabled,
                                                                       success: { cart in


                    // And now that we have a cart — time to redeem it.
                    strongSelf.registerDomainDetailsService.redeemCartUsingCredits(cart: cart,
                                                                                   domainContactInformation: strongSelf.jsonRepresentation(),
                                                                                   success: {

                            // Hey! We redeemed the cart sucessfully. Now we just need to set the new domain to primary...
                            strongSelf.registerDomainDetailsService.changePrimaryDomain(siteID: strongSelf.site.siteID,
                                                                             newDomain: strongSelf.domain.domainName,
                                                                             success: {

                                // We've succeeded! The domain is purchased and set to the primary one — time to drop out of this flow
                                // and return control to the Plugins, which will present the AT flow. (The VC handles that after getting the `.registerSucceeded` message.)
                                WPAnalytics.track(.automatedTransferCustomDomainPurchased)

                                strongSelf.isLoading = false
                                strongSelf.onChange?(.registerSucceeded(domain: strongSelf.domain.domainName))
                            },
                            failure: { error in

                                // This means we've sucessfully bought/redeemed the domain, but something went wrong
                                // when setting it to a primary.

                                strongSelf.isLoading = false
                                strongSelf.onChange?(.prefillError(message: Localized.changingPrimaryDomainError))
                        })
                    }, failure: { (error) in

                        // Failure during the purchase step. Not much we can do from the mobile in any case,
                        // so let's just show a generic error message.
                        WPAnalytics.track(.automatedTransferCustomDomainPurchaseFailed)
                        strongSelf.isLoading = false
                        strongSelf.onChange?(.prefillError(message: Localized.redemptionError))
                    })
            }) { (error) in

                // Same as above. If adding items to cart fails, not much we can do to recover :(
                WPAnalytics.track(.automatedTransferCustomDomainPurchaseFailed)
                strongSelf.isLoading = false
                strongSelf.onChange?(.prefillError(message: Localized.redemptionError))
            }
        })
    }

    func selectCountry(at index: Int) {
        let section = sections[SectionIndex.contactInformation.rawValue]
        if let row = section.rows[CellIndex.ContactInformation.country.rawValue].editableRow,
            let country = countries?[safe: index] {
            row.idValue = country.code
            row.value = country.name
            fetchStates(countryCode: country.code)
        }
    }

    func selectState(at index: Int) {
        if let state = states?[safe: index] {
            stateRow?.idValue = state.code
            stateRow?.value = state.name
        }
    }

    private func clearStateSelection() {
        stateRow?.idValue = nil
        stateRow?.value = nil
    }

    private var stateRow: EditableKeyValueRow? {
        let section = sections[SectionIndex.address.rawValue]
        return section.rows[safe: addressSectionIndexHelper.stateIndex]?.editableRow
    }

    func prefill() {
        fetchCountries { [weak self] in
            self?.fetchDomainContactInformation()
        }
    }

    private func fetchDomainContactInformation() {
        isLoading = true
        registerDomainDetailsService.getDomainContactInformation(
            success: { [weak self] (domainContactInformation) in
                guard let strongSelf = self else {
                    return
                }

                defer {
                    strongSelf.isLoading = false
                }

                let prefillSuccessBlock = {
                    strongSelf.update(with: domainContactInformation)
                    strongSelf.onChange?(.prefillSuccess)

                }
                if let countryCode = domainContactInformation.countryCode {
                    strongSelf.fetchStates(countryCode: countryCode) {
                        prefillSuccessBlock()
                    }
                } else {
                    prefillSuccessBlock()
                }
        }) { [weak self] (error) in
            guard let strongSelf = self else {
                return
            }
            strongSelf.isLoading = false
            strongSelf.onChange?(.prefillError(message: Localized.prefillError))
        }
    }

    private func fetchCountries(successCompletion: @escaping () -> Void) {
        isLoading = true
        registerDomainDetailsService.getSupportedCountries(success: { [weak self] (countriesResponse) in
            guard let strongSelf = self else {
                return
            }

            defer {
                strongSelf.isLoading = false
            }

            var result: [CodeNameTuple] = []
            //Filter empty records
            for country in countriesResponse {
                if let code = country.code,
                    let name = country.name,
                    !code.isEmpty,
                    !name.isEmpty {
                    result.append((code: code, name: name))
                }
            }
            strongSelf.countries = result
            successCompletion()
        }) { [weak self] (error) in
            guard let strongSelf = self else {
                return
            }
            strongSelf.isLoading = false
            strongSelf.onChange?(.prefillError(message: Localized.prefillError))
        }
    }

    private func fetchStates(countryCode: String, successCompletion: (() -> Void)? = nil) {
        isLoading = true
        states = nil
        clearStateSelection()
        registerDomainDetailsService.getStates(
            for: countryCode,
            success: { [weak self] (statesResponse) in
                guard let strongSelf = self else {
                    return
                }
                strongSelf.isLoading = false
                var result: [CodeNameTuple] = []
                //Filter empty records
                for state in statesResponse {
                    if let code = state.code,
                        let name = state.name,
                        !code.isEmpty,
                        !name.isEmpty {
                        result.append((code: code, name: name))
                    }
                }
                strongSelf.states = result
                successCompletion?()
        }) { [weak self] (error) in
            guard let strongSelf = self else {
                return
            }
            strongSelf.isLoading = false
            strongSelf.onChange?(.unexpectedError(message: Localized.statesFetchingError))
        }
    }

    private func update(with domainContactInformation: DomainContactInformation) {
        updateAddressSection(with: domainContactInformation)
        updatePhoneSection(with: domainContactInformation)
        updateContactInformationSection(with: domainContactInformation)
    }

    private func updateAddressSection(with domainContactInformation: DomainContactInformation) {
        let section = sections[SectionIndex.address.rawValue]
        section.rows[safe: addressSectionIndexHelper.cityIndex]?.editableRow?.value = domainContactInformation.city
        section.rows[safe: addressSectionIndexHelper.postalCodeIndex]?.editableRow?.value = domainContactInformation.postalCode
        section.rows[safe: addressSectionIndexHelper.addressLine1]?.editableRow?.value = domainContactInformation.address1
        section.rows[safe: addressSectionIndexHelper.stateIndex]?.editableRow?.idValue = domainContactInformation.state
        section.rows[safe: addressSectionIndexHelper.stateIndex]?.editableRow?.value = states?.filter {
            return $0.code == domainContactInformation.state
            }.first?.name
    }

    private func updatePhoneSection(with domainContactInformation: DomainContactInformation) {
        let section = sections[SectionIndex.phone.rawValue]
        if let phone = domainContactInformation.phone {
            let phoneNumberParts = phone.replacingOccurrences(of: Constant.phoneNumberCountryCodePrefix, with: "").split(separator: Constant.phoneNumberConnectingChar)
            if phoneNumberParts.count == 2 {
                section.rows[safe: CellIndex.PhoneNumber.countryCode.rawValue]?.editableRow?.value = String(phoneNumberParts[safe: 0] ?? "")
                section.rows[safe: CellIndex.PhoneNumber.number.rawValue]?.editableRow?.value = String(phoneNumberParts[safe: 1] ?? "")
            }
        }
    }

    private func updateContactInformationSection(with domainContactInformation: DomainContactInformation) {
        let section = sections[SectionIndex.contactInformation.rawValue]
        section.rows[safe: CellIndex.ContactInformation.country.rawValue]?.editableRow?.idValue = domainContactInformation.countryCode
        section.rows[safe: CellIndex.ContactInformation.country.rawValue]?.editableRow?.value = countries?.filter {
            return $0.code == domainContactInformation.countryCode
            }.first?.name
        section.rows[safe: CellIndex.ContactInformation.email.rawValue]?.editableRow?.value = domainContactInformation.email
        section.rows[safe: CellIndex.ContactInformation.firstName.rawValue]?.editableRow?.value = domainContactInformation.firstName
        section.rows[safe: CellIndex.ContactInformation.lastName.rawValue]?.editableRow?.value = domainContactInformation.lastName
        section.rows[safe: CellIndex.ContactInformation.organization.rawValue]?.editableRow?.value = domainContactInformation.organization
    }

    private func jsonRepresentation() -> [String: String] {
        var dict: [String: String] = [:]
        if let privacySectionSelectedItem = privacySectionSelectedItem() {
            dict[privacySectionSelectedItem.jsonKey] = String(privacySectionSelectedItem.rawValue)
        }
        dict.merge(phoneNumberSectionJson(), uniquingKeysWith: { (first, _) in first })
        dict.merge(sectionJson(sectionIndex: .contactInformation), uniquingKeysWith: { (first, _) in first })
        dict.merge(sectionJson(sectionIndex: .address), uniquingKeysWith: { (first, _) in first })
        return dict
    }

    private func privacySectionSelectedItem() -> CellIndex.PrivacyProtection? {
        let privacySection = sections[SectionIndex.privacyProtection.rawValue]
        for (index, row) in privacySection.rows.enumerated() {
            switch row {
            case .checkMark(let checkMarkRow):
                if checkMarkRow.isSelected {
                    return CellIndex.PrivacyProtection(rawValue: index)
                }
            default:
                break
            }
        }
        return nil
    }

    private func sectionJson(sectionIndex: SectionIndex) -> [String: String] {
        var dict: [String: String] = [:]
        let section = sections[sectionIndex.rawValue]
        for row in section.rows {
            switch row {
            case .inlineEditable(let editableRow):
                dict[editableRow.jsonKey] = editableRow.jsonValue
            default:
                break
            }
        }
        return dict
    }

    private func phoneNumberSectionJson() -> [String: String] {
        let section = sections[SectionIndex.phone.rawValue]
        let jsonKey = section.rows[CellIndex.PhoneNumber.number.rawValue].editableRow?.jsonKey ?? ""
        var dict: [String: String] = [:]
        dict[jsonKey] = formattedPhoneNumber()
        return dict
    }

    private func formattedPhoneNumber() -> String {
        let section = sections[SectionIndex.phone.rawValue]
        let countryCode = section.rows[CellIndex.PhoneNumber.countryCode.rawValue].editableRow?.value ?? ""
        let number = section.rows[CellIndex.PhoneNumber.number.rawValue].editableRow?.value ?? ""
        return Constant.phoneNumberCountryCodePrefix + countryCode +
            String(Constant.phoneNumberConnectingChar) + number
    }
}

// MARK: - Validate remotely

extension RegisterDomainDetailsViewModel {

    fileprivate func validateRemotely(successCompletion: @escaping () -> Void) {
        registerDomainDetailsService.validateDomainContactInformation(
            contactInformation: jsonRepresentation(),
            domainNames: [domain.domainName],
            success: { [weak self] (response) in
                guard let strongSelf = self else {
                    return
                }

                strongSelf.isLoading = false

                if response.success {
                    strongSelf.clearValidationErrors()
                    successCompletion()
                } else {
                    WPAnalytics.track(.automatedTransferCustomDomainContactInfoValidationFailed)
                    strongSelf.updateValidationErrors(with: response.messages)
                }
                strongSelf.onChange?(.remoteValidationFinished)
        }) { [weak self] (error) in
            guard let strongSelf = self else {
                return
            }
            strongSelf.isLoading = false
            strongSelf.onChange?(.unexpectedError(message: Localized.unexpectedError))
        }
    }

    fileprivate func clearValidationErrors() {
        for section in sections {
            for row in section.rows {
                if let editableRow = row.editableRow {
                    editableRow.firstRule(
                        forContext: ValidationRule.Context.serverSide
                        )?.isValid = true
                }
            }
        }
    }

    fileprivate func updateValidationErrors(with messages: ValidateDomainContactInformationResponse.Messages?) {
        guard let messages = messages else {
            return
        }
        updateContactInformationValidationErrors(messages: messages)
        updatePhoneNumberValidationErrors(messages: messages)
        updateAddressSectionValidationErrors(messages: messages)
    }

    fileprivate func updatePhoneNumberValidationErrors(messages: ValidateDomainContactInformationResponse.Messages) {
        let rows = sections[SectionIndex.phone.rawValue].rows
        let isValid = messages.isValidPhoneNumber()
        for row in rows {
            row.editableRow?.firstRule(
                forContext: .serverSide
                )?.isValid = isValid
        }
    }

    fileprivate func updateContactInformationValidationErrors(messages: ValidateDomainContactInformationResponse.Messages) {
        let rows = sections[SectionIndex.contactInformation.rawValue].rows
        for (index, row) in rows.enumerated() {
            if let editableRow = row.editableRow,
                let cellIndex = CellIndex.ContactInformation(rawValue: index) {
                editableRow.firstRule(
                    forContext: .serverSide
                    )?.isValid = messages.isValid(for: cellIndex)
            }
        }
    }

    fileprivate func updateAddressSectionValidationErrors(messages: ValidateDomainContactInformationResponse.Messages) {
        let rows = sections[SectionIndex.address.rawValue].rows
        for (index, row) in rows.enumerated() {
            if let editableRow = row.editableRow {
                let addressField = addressSectionIndexHelper.addressField(for: index)
                editableRow.firstRule(
                    forContext: .serverSide
                    )?.isValid = messages.isValid(addressField: addressField)
            }
        }
    }
}

extension ValidateDomainContactInformationResponse.Messages {

    typealias ContactInformation = RegisterDomainDetailsViewModel.CellIndex.ContactInformation
    typealias AddressField = RegisterDomainDetailsViewModel.CellIndex.AddressField
    typealias PhoneNumber = RegisterDomainDetailsViewModel.CellIndex.PhoneNumber

    func isValid(for index: ContactInformation) -> Bool {
        switch index {
        case .country:
            return countryCode?.isEmpty ?? true
        case .email:
            return email?.isEmpty ?? true
        case .firstName:
            return firstName?.isEmpty ?? true
        case .lastName:
            return lastName?.isEmpty ?? true
        default:
            return true
        }
    }

    func isValidPhoneNumber() -> Bool {
        return phone?.isEmpty ?? true
    }

    func isValid(addressField: AddressField) -> Bool {
        switch addressField {
        case .addressLine:
            return address1?.isEmpty ?? true
        case .city:
            return city?.isEmpty ?? true
        case .postalCode:
            return postalCode?.isEmpty ?? true
        case .state:
            return state?.isEmpty ?? true
        default:
            return true
        }
    }
}
